from typing import Annotated, Literal

import httpx
from eth_rpc import PrivateKeyWallet
from eth_rpc.networks import get_network_by_name
from eth_typeshed import ERC20
from eth_typeshed.multicall import multicall
from typing_extensions import Doc

from emp_agents.models.protocol import SkillSet, onchain_action, view_action


class ERC20Skill(SkillSet):
    """
    Tools for interacting with ERC20 tokens.
    """

    @view_action
    @staticmethod
    async def describe_protocol():
        """Returns the complete protocol specification of the ERC20 protocol"""

        async with httpx.AsyncClient() as client:
            response = await client.get(
                "https://raw.githubusercontent.com/ethereum/ercs/refs/heads/master/ERCS/erc-20.md"
            )
            return response.text

    @view_action
    @staticmethod
    async def get_token_info(
        network: Annotated[
            Literal["ethereum", "arbitrum", "base"],
            Doc("The network to use.  One of ethereum, arbitrum, or base."),
        ],
        token_address: Annotated[str, Doc("The address of the ERC20 token.")],
    ) -> str:
        """Returns the name, symbol and decimals for an ERC20 token"""

        try:
            network_type = get_network_by_name(network)
        except ValueError:
            return "Invalid network"

        token = ERC20[network_type](address=token_address)

        try:
            (name, symbol, decimals) = await multicall[network_type].execute(
                token.name(), token.symbol(), token.decimals()
            )
        except Exception as e:
            return f"Error getting token info: {e}"

        return f"name: {name}; symbol: {symbol}; decimals: {decimals}"

    @view_action
    @staticmethod
    async def get_balance(
        network: Annotated[
            Literal["ethereum", "arbitrum", "base"],
            Doc("The network to use.  One of ethereum, arbitrum, or base."),
        ],
        token_address: Annotated[str, Doc("The address of the ERC20 token.")],
        address: Annotated[
            str, Doc("The address of the account to get the balance of.")
        ],
        decimals: Annotated[int, Doc("How many decimals the token has")] = 18,
    ) -> str:
        """Returns the balance of an account for an ERC20 token"""

        try:
            network_type = get_network_by_name(network)
        except ValueError:
            return "Invalid network"

        token = ERC20[network_type](address=token_address)
        balance = await token.balance_of(address).get()
        return f"Balance: {balance / 10 ** decimals}"

    @onchain_action
    @staticmethod
    async def transfer(
        network: Annotated[
            Literal["ethereum", "arbitrum", "base"],
            Doc("The network to use.  One of ethereum, arbitrum, or base."),
        ],
        token_address: Annotated[str, Doc("The address of the ERC20 token.")],
        private_key: Annotated[
            str, Doc("The private key of the account to transfer from.")
        ],
        to_address: Annotated[str, Doc("The address of the account to transfer to.")],
        amount: Annotated[float, Doc("The amount to transfer.")],
    ) -> str:
        try:
            network_type = get_network_by_name(network)
        except ValueError:
            return "Invalid network"

        token = ERC20[network_type](address=token_address)
        wallet = PrivateKeyWallet(private_key=private_key)
        tx = await token.transfer(to_address, amount).execute(wallet)
        return f"Transaction sent: {tx.hash}"

    @onchain_action
    @staticmethod
    async def approve(
        network: Annotated[
            Literal["ethereum", "arbitrum", "base"],
            Doc("The network to use.  One of ethereum, arbitrum, or base."),
        ],
        token_address: Annotated[str, Doc("The address of the ERC20 token.")],
        private_key: Annotated[
            str, Doc("The private key of the account to transfer from.")
        ],
        spender: Annotated[str, Doc("The address of the account to transfer to.")],
        amount: Annotated[float, Doc("The amount to transfer.")],
    ) -> str:
        try:
            network_type = get_network_by_name(network)
        except ValueError:
            return "Invalid network"

        token = ERC20[network_type](address=token_address)
        wallet = PrivateKeyWallet(private_key=private_key)
        tx = await token.approve(spender, amount).execute(wallet)
        return f"Transaction sent: {tx.hash}"

    @onchain_action
    @staticmethod
    async def transfer_from(
        network: Annotated[
            Literal["ethereum", "arbitrum", "base"],
            Doc("The network to use.  One of ethereum, arbitrum, or base."),
        ],
        token_address: Annotated[str, Doc("The address of the ERC20 token.")],
        private_key: Annotated[
            str, Doc("The private key of the account to transfer from.")
        ],
        from_address: Annotated[
            str, Doc("The address of the account to transfer from.")
        ],
        to_address: Annotated[str, Doc("The address of the account to transfer to.")],
        amount: Annotated[float, Doc("The amount to transfer.")],
    ) -> str:
        try:
            network_type = get_network_by_name(network)
        except ValueError:
            return "Invalid network"

        token = ERC20[network_type](address=token_address)
        wallet = PrivateKeyWallet(private_key=private_key)
        tx = await token.transfer_from(from_address, to_address, amount).execute(wallet)
        return f"Transaction sent: {tx.hash}"



from emp_agents.tools.dexscreener import DexScreenerSkill
from emp_agents.tools.protocol import ERC20Skill, GmxSkill, SimpleWalletSkill

__all__ = ["ERC20Skill", "SimpleWalletSkill", "GmxSkill", "DexScreenerSkill"]



from emp_agents.models.anthropic import AnthropicBase
from emp_agents.models.openai import OpenAIBase
from emp_agents.models.shared import Message, ModelType, Request, Role
from emp_agents.models.shared.tools import GenericTool, Property

__all__ = [
    "AnthropicBase",
    "GenericTool",
    "Message",
    "ModelType",
    "OpenAIBase",
    "Property",
    "Request",
    "Role",
]



import json
from datetime import datetime
from typing import Any

from pydantic import AliasChoices, BaseModel, Field, field_serializer
from pydantic.types import Json

from emp_agents.types.enums import Role


class Function(BaseModel):
    name: str
    arguments: Json[Any]

    @field_serializer("arguments")
    def serialize_dt(self, dt: datetime, _info):
        return json.dumps(self.arguments)


class ToolCall(BaseModel):
    id: str
    type: str
    function: Function

    def __repr__(self):
        return f"<ToolCall function={self.function}>"


class Message(BaseModel):
    role: Role
    content: str | None
    tool_call_id: str | None = Field(
        default=None, validation_alias=AliasChoices("tool_call_id")
    )  # given to user in response
    refusal: str | None = Field(default=None)
    tool_calls: list[ToolCall] | None = Field(default=None)  # only in response

    @classmethod
    def build(
        self,
        content: str,
        role: Role = Role.user,
        tool_call_id: str | None = None,
        tool_calls: list[Any] | None = None,
    ):
        return Message(
            content=content,
            role=role,
            tool_calls=tool_calls,
            tool_call_id=tool_call_id,
        )

    def serialize_anthropic(self):
        data: dict[str, Any] = {"role": self.role}
        data["content"] = [{"type": "text", "text": self.content}]
        return data

    def __repr__(self):
        if self.content:
            return f"{self.role.value}: {self.content}"
        return f"{self.role.value}: {self.tool_calls}"

    __str__ = __repr__



from emp_agents.tools.protocol.erc20 import ERC20Skill
from emp_agents.tools.protocol.gmx import GmxSkill
from emp_agents.tools.protocol.wallets.simple import SimpleWalletSkill

__all__ = [
    "ERC20Skill",
    "SimpleWalletSkill",
    "GmxSkill",
]



from asyncio import iscoroutine
from typing import Any

from pydantic.types import Json


async def execute_tool(tools_map, function_name: str, arguments: Json[Any]):
    print("EXECUTING TOOL", function_name, arguments)

    func = tools_map[function_name]
    response = func(**arguments)
    if iscoroutine(response):
        return await response
    return response



from typing import Any

from pydantic import Field

from emp_agents.models.protocol import SkillSet

from .base import AgentBase


class SkillsAgent(AgentBase):
    skills: list[type[SkillSet]] = Field(default_factory=list)

    def model_post_init(self, __context: Any):
        super().model_post_init(__context)

        for skill in self.skills:
            for tool in skill._tools:
                self._add_tool(tool)



from contextvars import ContextVar
from typing import Optional

from eth_rpc import PrivateKeyWallet

from emp_agents.models.protocol import SkillSet, tool_method

# Context variable to store private key
_private_key: ContextVar[Optional[str]] = ContextVar("_private_key", default=None)


class SimpleWalletSkill(SkillSet):
    """A simple wallet tool that stores private key in memory using context vars"""

    @tool_method
    @staticmethod
    def create_wallet() -> str:
        """Create a new private key wallet"""

        wallet = PrivateKeyWallet.create_new()
        _private_key.set(wallet.private_key)
        return (
            f"Wallet created: {wallet.address} with private key: {wallet.private_key}"
        )

    @tool_method
    @staticmethod
    def set_private_key(private_key: str) -> str:
        """Set the private key in the context"""

        _private_key.set(private_key)
        return "Private key set successfully"

    @tool_method
    @staticmethod
    def get_private_key() -> str:
        """Get the private key from the context"""

        key = _private_key.get()
        if key is None:
            return "No private key set"
        return key

    @tool_method
    @staticmethod
    def clear_private_key() -> str:
        """Clear the private key from the context"""

        _private_key.set(None)
        return "Private key cleared"

    @tool_method
    @staticmethod
    def get_address() -> str:
        """Get the address of the wallet"""

        key = _private_key.get()
        if key is None:
            return "No private key set"
        wallet = PrivateKeyWallet(private_key=key)
        return wallet.address



from textwrap import dedent
from typing import Any, Callable

from pydantic import BaseModel, Field, PrivateAttr, field_validator

from emp_agents.exceptions import InvalidModelException
from emp_agents.models import AnthropicBase, GenericTool, Message, OpenAIBase, Request
from emp_agents.types import AnthropicModelType, OpenAIModelType, Role
from emp_agents.utils import count_tokens, execute_tool, summarize_conversation


class AgentBase(BaseModel):
    agent_id: str = Field(default="")
    description: str = Field(default="")
    default_model: OpenAIModelType | AnthropicModelType | None = None
    prompt: str = Field(default="You are a helpful assistant")
    personality: str | None = Field(default=None)
    tools: list[GenericTool] = Field(default_factory=list)
    conversation_history: list[Message] = Field(default_factory=list)
    requires: list[str] = []
    openai_api_key: str | None = None
    anthropic_api_key: str | None = None

    _tools: list[GenericTool] = PrivateAttr(default_factory=list)
    _tools_map: dict[str, Callable[..., Any]] = PrivateAttr(default_factory=dict)

    @field_validator("prompt", mode="before")
    @classmethod
    def to_prompt(cls, v: str) -> str:
        return dedent(v).strip()

    @field_validator("tools", mode="before")
    @classmethod
    def to_generic_tools(
        cls, v: list[Callable[..., Any] | GenericTool]
    ) -> list[GenericTool]:
        return [
            GenericTool.from_func(tool) if not isinstance(tool, GenericTool) else tool
            for tool in v
        ]

    def model_post_init(self, _context: Any):
        for tool in self.tools:
            if isinstance(tool, GenericTool):
                self._tools.append(tool)
            else:
                self._tools.append(GenericTool.from_func(tool))

        self._tools_map = {tool.name: tool.func for tool in self._tools}
        self.conversation_history = [
            Message(role=Role.system, content=self.system_prompt)
        ] + self.conversation_history

    def get_token_count(
        self, model: OpenAIModelType | AnthropicModelType = OpenAIModelType.gpt4o_mini
    ) -> int:
        return count_tokens(self.conversation_history, model)

    async def summarize(
        self,
        model: OpenAIModelType | AnthropicModelType = OpenAIModelType.gpt4o_mini,
        update: bool = True,
    ) -> str:
        summary = await summarize_conversation(
            self._make_client(model), self.conversation_history, model=model
        )
        if update:
            self.conversation_history = [summary]
        assert summary.content is not None, "Summary content should always be present"
        return summary.content

    async def respond(
        self,
        question: str,
        model: OpenAIModelType | AnthropicModelType | None = None,
        max_tokens: int | None = None,
        response_format: type[BaseModel] | None = None,
    ) -> str:
        if self.default_model is None and model is None:
            raise InvalidModelException("Model is required")
        if model is None:
            assert self.default_model is not None, "Model is required"
            model = self.default_model
        if not isinstance(model, (OpenAIModelType, AnthropicModelType)):
            raise InvalidModelException(f"Invalid model: {model}")
        client: OpenAIBase | AnthropicBase = self._make_client(model)
        conversation = [
            Message(role=Role.system, content=self.system_prompt),
            Message(role=Role.user, content=question),
        ]

        while True:
            request = Request(
                messages=conversation,
                model=model,
                tools=self._tools,
                max_tokens=max_tokens or 1_000,
                response_format=response_format,
            )
            response = await client.completion(
                request,
            )

            if isinstance(model, OpenAIModelType):
                conversation += response.messages
            else:
                conversation += [Message(role=Role.assistant, content=response.text)]

            tool_calls = response.tool_calls

            if tool_calls:
                for tool_call in tool_calls:
                    result = await execute_tool(
                        self._tools_map,
                        tool_call.function.name,
                        tool_call.function.arguments,
                    )
                    message = Message(
                        role=(
                            Role.user
                            if isinstance(model, AnthropicModelType)
                            else Role.tool
                        ),
                        content=result,
                        tool_call_id=(
                            tool_call.id if isinstance(model, OpenAIModelType) else None
                        ),
                    )
                    conversation += [message]
                continue
            else:
                return response.text

    async def answer(
        self,
        question: str,
        model: OpenAIModelType | AnthropicModelType = OpenAIModelType.gpt4o_mini,
    ) -> str:
        self.conversation_history += [Message(role=Role.user, content=question)]
        response = await self.complete(
            model=model,
        )
        return response

    async def step(
        self,
        model: OpenAIModelType | AnthropicModelType = OpenAIModelType.gpt4o_mini,
    ) -> None:
        client: OpenAIBase | AnthropicBase = self._make_client(model)
        request = Request(
            messages=self.conversation_history, model=model, tools=self._tools
        )
        response = await client.completion(
            request,
        )
        if isinstance(model, OpenAIModelType):
            self.conversation_history += response.messages
        else:
            self.conversation_history += [
                Message(role=Role.assistant, content=response.text)
            ]

        tool_calls = response.tool_calls
        if tool_calls:
            for tool_call in tool_calls:
                result = await execute_tool(
                    self._tools_map,
                    tool_call.function.name,
                    tool_call.function.arguments,
                )
                message = Message(
                    role=(
                        Role.user
                        if isinstance(model, AnthropicModelType)
                        else Role.tool
                    ),
                    content=result,
                    tool_call_id=(
                        tool_call.id if isinstance(model, OpenAIModelType) else None
                    ),
                )
                self.conversation_history += [message]
        return None

    async def complete(
        self, model: OpenAIModelType | AnthropicModelType, verbose: bool = False
    ) -> str:
        client: OpenAIBase | AnthropicBase = self._make_client(model)

        while True:
            request = Request(
                messages=self.conversation_history, model=model, tools=self._tools
            )
            response = await client.completion(
                request,
            )

            if isinstance(model, OpenAIModelType):
                self.conversation_history += response.messages
            else:
                self.conversation_history += [
                    Message(role=Role.assistant, content=response.text)
                ]

            tool_calls = response.tool_calls

            if tool_calls:
                for tool_call in tool_calls:
                    result = await execute_tool(
                        self._tools_map,
                        tool_call.function.name,
                        tool_call.function.arguments,
                    )
                    message = Message(
                        role=(
                            Role.user
                            if isinstance(model, AnthropicModelType)
                            else Role.tool
                        ),
                        content=result,
                        tool_call_id=(
                            tool_call.id if isinstance(model, OpenAIModelType) else None
                        ),
                    )
                    if verbose:
                        print(message)
                    self.conversation_history += [message]
                continue
            else:
                return response.text

    def __repr__(self):
        prompt = self.prompt[:100].strip().replace("\n", " ")
        if len(prompt) >= 50:
            prompt = prompt[:50] + "..."
        return dedent(
            """
            <{class_name}
                prompt="{prompt}..."
                tools=[
                    {tools}
                ]
            >
        """.format(
                class_name=self.__class__.__name__,
                prompt=prompt,
                tools="\n".join([repr(tool) for tool in self.tools]),
            )
        ).strip()

    __str__ = __repr__

    def _make_client(
        self, model: OpenAIModelType | AnthropicModelType
    ) -> OpenAIBase | AnthropicBase:
        if isinstance(model, OpenAIModelType):
            if not self.openai_api_key:
                raise ValueError("OpenAI API key is required")
            return OpenAIBase(openai_api_key=self.openai_api_key)
        else:
            if not self.anthropic_api_key:
                raise ValueError("Anthropic API key is required")
            return AnthropicBase(anthropic_api_key=self.anthropic_api_key)

    async def __call__(
        self,
        question: str,
        model: OpenAIModelType | AnthropicModelType = OpenAIModelType.gpt4o_mini,
    ) -> str:
        return await self.answer(question, model)

    async def reset(self):
        self.conversation_history = []

    @property
    def system_prompt(self) -> str:
        prompt = self.prompt
        if self.personality:
            prompt = f"{prompt}\n\nPERSONALITY: {self.personality}"
        return prompt.strip()

    def print_conversation(self) -> None:
        for message in self.conversation_history:
            if not message.tool_call_id:
                print(f"{message.role}: {message.content}")

    def _make_message(self, content: str, role: Role = Role.user):
        return Message(role=role, content=content)

    async def run(self):
        conversation = [Message(role=Role.system, content=self.system_prompt)]
        while True:
            question = input("You: ")
            if question == "":
                break
            conversation += [Message(role=Role.user, content=question)]
            response = await self.answer(question)
            conversation += [Message(role=Role.assistant, content=response)]
            print(response)

    def _add_tool(self, tool: GenericTool) -> None:
        self._tools.append(tool)
        self._tools_map[tool.name] = tool.func
